"""General models provided for extension"""

from jbutils.models.attr_dict import AttrDict, AttrList
from jbutils.models.base import Base
from jbutils.models.console_theme import ConsoleTheme


__all__ = [
    "AttrDict",
    "AttrList",
    "Base",
    "ConsoleTheme",
]
